using System;
using System.Diagnostics;
using System.IO.Ports; 
using System.IO;
using System.Windows.Forms;
using System.Text;

using AddimatDeviceManager;

namespace AddimatVirtualComPort
{

	public class VirtualComPort : System.Windows.Forms.NativeWindow 
	{ 
	    
		public enum TransmitMode {
			Standard,
			Polling_ENQ1, 
			Polling_ENQ9,
			Polling_Alternate
		}

		public const String UsbDeviceName = "Addimat USB 2 Waiter Lock (serial)";

		public delegate void ReadEventHandler(TransmitMode mode, String data);
		public event ReadEventHandler PortReadEvent;

		public delegate void DisconnectedEventHandler(String port);
		public event DisconnectedEventHandler PortDisconnectedEvent;

		public delegate void ReconnectedEventHandler(String port);
		public event ReconnectedEventHandler PortReconnectedEvent;

		public delegate void ReconnectingEventHandler(String port);
		public event ReconnectingEventHandler PortReconnectingEvent; 
    
		public static System.IO.Ports.SerialPort gComPort; 
	

		private const int WM_DEVICECHANGE = 537;    // Change of a hardware configuration of a device or the computer. 
		private const int DBT_DEVNODES_CHANGED = 7; // A Device has been added or removed. 

		private bool gDeviceRemoved = false;
    
		protected override void WndProc(ref Message m) 
		{ 
			if (m.Msg == WM_DEVICECHANGE) { 
				this.HandleDeviceChange(ref m); 
			} 
			base.WndProc(ref m); 
		} 

    
		private void HandleDeviceChange(ref Message m) 
		{ 
			// usbser.sys does not report a broken (cable unplugged) port connection. 
			// There exists only the information that an USB port has changed (WParam=DBT_DEVNODES_CHANGED). 

			// Do nothing, when gComPort does not exist (should never occur).
			if (gComPort == null) return;
			if (gComPort.PortName == null) return;
			if (gComPort.PortName.Length == 0) return;

			String ConnectedComPorts;

			// This message occures on each device change, for example, when a memory stick has been removed.
			if ((int)m.WParam == DBT_DEVNODES_CHANGED)
			{
				// Scan current active COM port connections of the specified UsbDevice.
				ConnectedComPorts = DeviceManager.GetConnectedComPorts(UsbDeviceName);

				// gComPort.portNames contains the current port name that we use.
				// When gComPort.portNames is not in ConnectedComPorts, then the line has broken.
				if (ConnectedComPorts.Contains(gComPort.PortName) == false)
				{ 
					// We notice, that the connection has been broken, for further use.
					gDeviceRemoved = true;
					if (PortDisconnectedEvent != null) {
						// Inform the application about the event.
						PortDisconnectedEvent(gComPort.PortName);
					}
				}
				// When gComPort.portNames is in ConnectedComPorts, then the message was either from another
				// device (a memory stick, for example) or the connection to gComPort.portNames is not broken.
				else
				{
					// Whether the connection has been broken before, we can test by checking gDeviceRemoved.
					// If so, we have to reset the attached usbser.sys driver and to reopen the COM port.
					// Using the flag gDeviceRemoved is also necessary, because the device manager can report 
					// the same event more than once.
					if (gDeviceRemoved == true)
					{
						String PortName = gComPort.PortName; // Save current port name to be processed.
						gDeviceRemoved = false;

						// Inform the application about the event.
						if (PortReconnectingEvent != null)
						{
							PortReconnectingEvent(PortName);
						}
						try {
							gComPort.Close();
						}
						catch (Exception ex) {
							// Here, an exception will be fired, because the connection is broken.
							// but usbser.sys closes the broken connection regardless.
							Debug.WriteLine("Dev Handler Changed: Error on Closing Com Port "+ex.Message);
						}
						// Reset the appropriate driver usbser.sys by disabling and re-enabling.
						String FriendlyUsbPortName= UsbDeviceName + " (" + PortName + ")";
						DeviceManager.EnableDeviceState(FriendlyUsbPortName, false); // Disable
						DeviceManager.EnableDeviceState(FriendlyUsbPortName, true);  // Enable
						System.Threading.Thread.Sleep(200); 
						
						// Re-scan COM Port connections.
						// In fact, it should not to be neccessary to rescan the port connections, but
						// we have reset the connection since last port scan.
						ConnectedComPorts = DeviceManager.GetConnectedComPorts(UsbDeviceName);
						if (ConnectedComPorts.Contains(PortName) == true)
						{
              string errMsg = "";
              if (OpenComPort(PortName,ref errMsg) == true)
							{
								if (PortReconnectedEvent != null)
								{
									PortReconnectedEvent(PortName);
								}
							}
						}
					}
				}
			}
		}

		


		public VirtualComPort() 
		{ 
			this.CreateHandle(new CreateParams()); 
		} 
    
		~VirtualComPort()
		{
			this.DestroyHandle(); 
		}
	   
		

		public bool OpenComPort(string portName, ref string errorMsg) 
		{ 
      try {
				gComPort = new SerialPort(portName, 115200, Parity.None, 8, StopBits.One);
        gComPort.Open();
        gComPort.ReadTimeout = 2000;

        //----------------------------------------------------------------------//
        // Without DtrEnable = true, the USB Device will not transfer any data! //
        //----------------------------------------------------------------------//
        gComPort.DtrEnable = true;  
				
        errorMsg = "";
				return true;
			} 
			catch (Exception ex) {
				Debug.WriteLine("VirtualComPort OpenComPort Exception: " + portName+ " " + ex.Message);
                errorMsg = ex.Message;
				return false;
			} 
		} 

		public void CloseComPort() 
		{
			if (gComPort == null) return;
			try { 
				if (gComPort.IsOpen) { 
					gComPort.Close();
					System.Threading.Thread.Sleep(200);
					Debug.WriteLine("VirtualComPort CloseComPort: " + gComPort.PortName);
				}
			} 
			catch (Exception ex) {
				Debug.WriteLine("VirtualComPort CloseComPort Exception: " + gComPort.PortName + " " + ex.Message);
			} 
		} 

		private void StandardModeReading() 
		{ 
			try {
				int count = gComPort.BytesToRead;
				byte[] buffer = new byte[count];
				if ((count > 0)) { 
					gComPort.Read(buffer, 0, count);
					string result = System.Text.Encoding.ASCII.GetString(buffer); 
					if (PortReadEvent != null) {
						PortReadEvent(TransmitMode.Standard, result); 
					} 
				} 
			} 
			catch (Exception ex) {
				Debug.WriteLine("VirtualComPort StandardMode Exception: " + gComPort.PortName + " " + ex.Message);
				throw new Exception("VirtualComPort Standard Mode Exception: " + ex.Message); 
			} 
		} 

		private void PollingModeReading(TransmitMode enq) 
		{ 
			try { 
				switch (enq) {
					case TransmitMode.Polling_ENQ1:
						gComPort.Write("1");
						break;
					case TransmitMode.Polling_ENQ9:
						gComPort.Write("9");
						break;
					default: // Should never occur.
						return;
				}
				System.Threading.Thread.Sleep(200); 
				string result = ""; 
				int count = gComPort.BytesToRead; 
				if ((count > 0)) { 
					byte[] buffer = new byte[count];
					string tmp; 
					gComPort.Read(buffer, 0, count);
					for (int n = 0; n <= count - 1; n++) {
						tmp = buffer[n].ToString("X"); 
						if (tmp.Length == 1) { 
							result += "0" + tmp + " "; 
						} 
						else { 
							result += tmp + " "; 
						} 
					} 
				}
				if (PortReadEvent != null) {
					PortReadEvent(enq, result);
				}  
			} 
			catch (Exception ex) {
				Debug.WriteLine("VirtualComPort PollingMode Exception: " + gComPort.PortName + " " + ex.Message);
				throw new Exception("VirtualComPort Polling Mode Exception: " + ex.Message); 
			} 
      
		}

		private bool SendData(String text)
		{
			try
			{
				if (gComPort == null) return false;
				if (gComPort.IsOpen == false) return false;
				gComPort.Write(text);
				return true;
			}
			catch (Exception ex)
			{
				Debug.WriteLine("VirtualComPort SendData Exception: " + ex.Message);
			}
			return false;
		}

		public void RequestData(VirtualComPort.TransmitMode readingMode) 
		{ 
			try {
				if ((gComPort == null) == true) return;
				if (gComPort.IsOpen == false) return;
				switch (readingMode) {
					case VirtualComPort.TransmitMode.Standard:
						StandardModeReading(); 
						break;
					case VirtualComPort.TransmitMode.Polling_ENQ1:
					case VirtualComPort.TransmitMode.Polling_ENQ9:
						PollingModeReading(readingMode);
						break;
					default: // Should never occur
						break;
				} 
			} 
			catch (Exception ex) { 
				throw new Exception("VirtualComPort GetData Exception: " + ex.Message); 
			} 
		}   
	} 
}
